/*
  Macro           DirList
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Version         v4   28 Jan 2024
  Compatibility   Windows TSE v4 upwards


  This tool creates a list of directories and/or files.
  Directories are listed with their sizes!


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there,
  for example by opening it there in TSE and applying TSE's Macro Compile menu.


  INPUT

  To run the tool, execute it as a macro, for example from TSE's Macro Execute
  menu, or by adding it to TSE's Potpourri menu, or by assigning it to a key.

  When running it you can provide parameters after the macro name or at its
  prompt.

  Parameters can be directories and/or options.

  Directories can also be drives, or shares that start with "\\".
  Note that unlike Windows, the tool interprets "C:" as "C:\".

  The placement and order of options do not matter.
  Default the tool will list everything it can.
  This can be changed with these options:
    -d      Only list directories.
    -f      Only list files.
            List both.
    -d -f   List both.

  Listing directories default calculates their sizes.
  Calculating directory sizes makes the tool roughly twice as slow.
  You can list directories without calculating their sizes with the parameter:
    --nodirsizes


  OUTPUT

  The output will be sorted on the objects' full name.

  Each line represents an object; a directory or file.

  Lines have aligned fields that contain the object's properties:
    Last modified date    yyyy-mm-dd   A sortable format.
    Last modified time    hh:mm:ss     Seconds always even; a TSE limitation.
    Size                  Not limited to TSE's maximum integer size.
    Accessibility         1 = accessible by TSE, 0 = not accessible by TSE.
    Object type           d = directory, f = file.
    Full name             If necessary converted to TSE's character set.

  Field alignment is based on their actual sizes, so field positions might not
  always be the same.



  TODO

    MUST
    - .
    SHOULD
    - Dirlist v4.1:
      - For parameter "-x" generate Excel output:
        - Separate output fields with ";".
          My plan is to retain the fields' fixed position and width.
        - Quote non-numeric fields.
        - Make the extension .csv.
    - Dirlist v4.2:
      - Make digit grouping and date and time formats configurable.
        For digit grouping I will try to make the regional format the default.
        For dates and times the sortable format will remain the default,
        except perhaps when the "-x" option is used.
    - DirList v5:
      - I plan to research if DirList can use Windows APIs to also retrieve
        - long paths,
        - files with non-TSE characters in their names,
        - alternate data streams (ADS) of Windows NTFS,
        - create and access date-times of directories and files,
        - partial seconds of time fields.
    COULD
    - Slowing down when accessing remote drives.
      Remote drives are already slow by themseves.
      Slowing their usage too might still be a good idea for two reasons:
      (1) In the past, when I tried to list a network drive, at some point it
          aborted on an error. My assumption was and is that it could not keep
          up with the requests.
      (2) If you share a remote drive with others, then you might not want to
          use it so intensely that others get bad response-times or worse.
      Because I ran the list after working hours, (1) was my main problem.
      But currently I have neither such remote drives, and therefore not a
      reproduceable problem.
      So I will postpone this point until I have a reproduceable problem,
      or until someone else reports having it.
    WONT
    - Protection against circular references.
      My theoretical worst-case scenario is a Windows computer having access to
      a Linux drive that contains a circular directory reference.
      A circular directory reference would put DirList in an infinite loop.
      I currently see no way to detect circular directory references from
      Windows.
      In Linux I myself have no practical use case.
      Besides, Linux is currently not supported anyway.


  LINUX COMPATIBILITY

  Linux compatibility is currently not possible because of a TSE bug or a Linux
  difference. See the "technical background info" section for details.
  The tool compiles in Linux, but when run it is likely to abort TSE on a
  "segmentation fault". Probably as a side effect of that, I have also seen it
  turn the terminal blank: type "reset" if that happens.


  EXCEPTIONS

  There are four known cases in which a directory or file is not listed or
  wrongly listed. They are explained in more detail further on.
  Here is a short overview:
  - If the full name of a parent directory has a too long path, then the child
    directories and files will not be listed.
  - If the name of a directory or file contains ANSI-incompatible characters,
    then Windows converts such characters to their closest ANSI equivalent,
    and shows TSE the converted but therefore wrong name.
    For example, the Greek character for "alpha" becomes an "a".
  - If the full name of a parent directory contains ANSI-incompatible
    characters, then the child directories and files will not be listed.
  - 64-bit Windows systems hide some system directories and files from 32-bit
    programs like TSE, for example "C:\Windows\System32\drivers\etc\hosts".

  In the first 3 cases this tool mitigates these flaws by indicating such
  directories and files or their parent directories are inaccessible.
  The 4th case this tool cannot see, so it does not list its occurrences.


  Too Long Paths

  Because TSE's string limit is 255 characters and because TSE uses the old
  Windows ANSI APIs which are likewise string limited, TSE can list the
  contents of a folder if its full path has no more than 255 characters.
  The first consequence is, that such a folder itself is inaccessible to TSE,
  and that all its contents are invisible to TSE.
  The second consequence is, that TSE can still "see" the name of a child
  folder or file with a full name of more than 255 characters if the full name
  of their parent folder is less than 255 characters, but that TSE can not
  access such child folders and files.


  Ansi-incompatible Characters In Name  (case 2 and 3)

  Windows folder and file names are stored using UTF-16LE character encoding,
  which supports 149,813 characters as of Unicode 15.1.
  Windows GUI TSE and Linux TSE default support ANSI (a.k.a. "Windows-1252")
  character encoding, which has 218 displayable characters.
  Windows Console TSE supports "code pages", which consist of 128 ASCII
  characters plus 128 code page-specific charcters.
  Both Windows TSEs use the old Windows ANSI APIs to access Windows folders and
  files.
  This has the benefit that if a Unicode character has an ANSI or code page
  equivalent, then the Windows API converts characters back and forth if
  possible.
  For instance, the Unicode character "latin small letter e with acute" exists
  in ANSI too, and the Windows API makes TSE think it reads and write the ANSI
  character for folder and file names.
  For Unicode characters without a TSE equivalent the Windows API converts
  the Unicode character one-way to a close TSE alternative.
  For instance, in folder and file names TSE sees the Greek letter alpha and
  other close "foreign" (to ANSI) characters as an "a".
  But having seen that, TSE can not open such a folder or file because its
  actual name does not contain an "a", and the low-level Windows API does not
  know, remember, or want to guess which of the foreign character possibilities
  that "a" came from.
  So, for Windows folder and file names that contain a Unicode character that
  has no two-way TSE equivalent DirList will show the close alternative
  character and mark the folder or file in the list as "inaccessible".


  Invisible 64-bit Folders And Files

  This problem only exists on 64-bit systems.
  For backwards compatibility both 32-bit and 64-bit programs can and
  should access "C:\Windows\System32" to get to Windows programs and files
  that perform Windows tasks.
  But typically those programs and files differ for 32-bit and 64-bit.
  Windows solves this as follows:
  - It maintains two folders:
    "C:\Windows\System32" for 64-bit programs and
    "C:\Windows\SysWOW64" for 32-bit programs.
  - Whenever a 32-bit program (like TSE) thinks it accesses the folder
    "C:\Windows\System32" Windows invisibly substitutes it with the
    folder "C:\Windows\SysWOW64".
  Presumably because Windows itself is 64-bit, folders and files that
  should only be relevant to Windows are only maintained in
  "C:\Windows\System32", but there they are invisible for 32-bit programs
  like TSE.
  DirList runs from TSE, so it is 32-bit too. It does not even see any 64-bit
  folders and files, so it can not even report them as "inaccessible".

  Example:
  When TSE opens "C:\" it cannot browse to the file
    "C:\Windows\System32\drivers\etc\hosts",
  because TSE is a 32-bit program, so for TSE the directory
    "etc"
  does not exist in the
    "C:\Windows\SysWOW64\drivers"
  directory.
  If the full filename is supplied, then Windows lets TSE open but not write
  the file.
  If TSE is run as Administrator AND the full filename is supplied,
  then TSE can open and write the file.
  But as you can see, there is no way to list a "64-bit" parent directory.


  HISTORY

  v1         23 Dec 1999
    Initial release.

  v2          8 Jul 2004
    Starting from commandline didn't work correctly anyway, so deleted that.
    Now also accepts quoted dirs in parameter.
    Added parameter "all" for all files in all directories.
    <Escape> now interrupts the running macro.
    Added windmill to still show progress during long path processing.
    Added file-date, -time and -size.
    Known bug: Shown file size is erroneous above 2 GB.

  v3         24 Aug 2019
    Modernized the source code format to my newer standards.
    Now does not abort on but reports the top folders and files that TSE
    cannot access:
      With too deep paths ( > 255 characters).
      With ANSI-incompatible names.
    As "top" implies, the contents of inaccessible folders remain unlisted.

  v3.0.1     13 Oct 2019
    Corrected a textual imperfection in the final result message.
    Documented why folders and files with deep or Unicode path/names are
    inaccessible.
    Documented why 64-bit files in "C:\Windows\System32" are not listed.

  v3.0.2      3 Dec 2019
    Tweaked the documentation.

  v3.1        8 Jan 2020
    Show file sizes with grouped digits.
    Show the correct file size for files larger than 2,147,483,647 bytes
    for TSE versions from v4.40.95 upwards.

  v3.2   25 Jan 2024
  - Added a sound and final statistics pop-up when ready, both because it is
    user-friendlier and to make DirList better comparable to what is currently
    DirList v4, the next major version of DirList that is in development.
  - Included the current date/time in the generated file's name.

  v4   28 Jan 2024
  - The tool was rewritten to fix the recursion bug for too many subdirectories.
  - This does not fix too long paths and names: That is a different problem.
  - Directories were added to DirList, making the DirSize tool obsolete.
  - You can still list directories or files separately by using parameters.
  - Object type is now a column: "d" or "f" for a directory or file.
  - Accessibility is now a column: "1" or "0" for accessible or not accessible
    by TSE.

*/






/*
  T E C H N I C A L   B A C K G R O U N D   I N F O


  NOTE

    This is DirList v4.
    It does not support alternate data streams (ADS) yet.
    However, anticipating supporting ADS in DirList v5,
    it is sometimes already mentioned in descriptions and code.


  CONTEXT

    One user reported a bug, which turned out to be caused by recursively
    processing a file system with so many levels of subdirectories,
    that the limit of TSE's stack was reached, which I belief to be 64 KiB.

    Macro procedures can typically cost an unexpected amount of stack space
    when they use TSE statements that use strings.
    TSE statements do not store their internal variables on the stack at
    run-time for the duration of the statement, but the internal variables'
    sizes is already allocated at compile-time as if they were one of the
    procedure's variables.
    According to Semware this was done because it greatly simplifies
    bookkeeping.
    I suspect, that in general it makes macros faster.
    Its downside is, that it limits those macros that use recursion.

    Note:
      Generating a macro's .map file with "sc32 -m ..." shows each of its
      procedure's stack sizes.

    Because of the user's bug DirList v4 was rewritten to not use recursion.


  LIST A DIRECTORY'S SUBDIRECTORIES AND FILES WITHOUT RECURSION

    We start with a buffer that lists the input parameters' directories.
    This list is processed top to bottom line by line.
    If a line represents a directory, then its child objects are added as lines
    directly below it.

    As the list is being generated, all the normal directory and file
    properties are added immediately, with the following exceptions and
    additions:
    - A directory's size is initially set to 0.
    - A temporary "parent line" field is added, pointing to the line of the
      parent directory or file. (An ADS's parent is a directory or a file.)
    - A temporary "group size" field is added.
      For a directory it is set to 0.
      For a file and ADS it is set to their own size, and added to all their
      ancestors' group sizes.

    What remains for a next stage is
    - copying directory group sizes to directory sizes,
    - deleting the columns for "parent line" and "group size",
    - and sort and format the list as desired.


  SLOWNESS OF CALCULATING DIRECTORY SIZES

  The outer part of the algorithm is walking the tree.
  I think this is already done efficiently.

  The middle part of the algorithm is adding group sizes to all their
  ancestors.
  This requires updating some ancestors multiple times.
  I think that this too is already done efficiently.
  Past experiences suggest that updating lines can be relatively slow,
  but by this much?
  Something to try to test alternatives for, but I am not hopeful.

  The inner part of the algorithm is the add_bigints() procedure.
  I expect that add_bigints() can be optimized, maybe in two ways.
  - It currently processes digits one by one, which is inefficient.
    I have hopes that this can be inproved significantly.
  - It uses sub-procs which copy parameters.
    Not doing both (if possible) might improve performance a little.


  LINUX INCOMPATIBILITY

  In Windows TSE's FindFirstFile()/FindNextFile() command can list a directory
  that has a path equal to or shorter than 253 bytes, even if one of its child
  objects would have a full path that is longer than 255 bytes.
  The tool can then pre-calculate the too long length of the child object and
  avoid accessing it further.
  In Linux the same FindFirstFile()/FindNextFile() command aborts TSE with a
  "segmentation fault".

*/





// Compatibility restrictions and mitigations

/*
  When compiled with a TSE version below TSE 4.0 the compiler reports
  a syntax error and hilights the first applicable line below.

  There is a beta Linux version of TSE that is not bug-free and in which some
  significant features do not work, but all its Linux versions are above
  TSE 4.0, and they all are 32-bits which is what WIN32 actually signifies.
*/

#ifdef LINUX
  #define WIN32 FALSE
#endif

#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifndef EDITOR_VERSION
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4500h

  // A few intermediate TSE versions after v4.40 and before v4.40.95
  // are not supportable and will raise a compiler error on FFSizeStr().

  string proc FFSizeStr()
    return(Str(FFSize()))
  end FFSizeStr
#endif



#if EDITOR_VERSION < 4400h
  /*
    StrReplace() 1.0

    If you have TSE Pro 4.0 or 4.2, then this proc almost completely implements
    the built-in StrReplace() function of TSE Pro 4.4.
    The StrReplace() function replaces a string (pattern) inside a string.
    It works for strings like the Replace() function does for files, so read
    the Help for the Replace() function for the usage of the options, but apply
    these differences while reading:
    - Where Replace() refers to "file" and "line", StrReplace() refers to
      "string".
    - The options "g" ("global", meaning "from the start of the string")
      and "n" ("no questions", meaning "do not ask for confirmation on
      replacements") are implicitly always active, and can therefore be omitted.
    Notable differences between the procedure below with TSE 4.4's built-in
    function are, that here the fourth parameter "options" is mandatory
    and that the fifth parameter "start position" does not exist.
  */
  integer strreplace_id = 0
  string proc StrReplace(string needle, string haystack, string replacer, string options)
    integer i                      = 0
    integer org_id                 = GetBufferId()
    string  result  [MAXSTRINGLEN] = haystack
    string  validated_options [20] = 'gn'
    for i = 1 to Length(options)
      if (Lower(SubStr(options, i, 1)) in '0'..'9', 'b', 'i','w', 'x', '^', '$')
        validated_options = validated_options + SubStr(options, i, 1)
      endif
    endfor
    if strreplace_id == 0
      strreplace_id = CreateTempBuffer()
    else
      GotoBufferId(strreplace_id)
      EmptyBuffer()
    endif
    InsertText(haystack, _INSERT_)
    lReplace(needle, replacer, validated_options)
    result = GetText(1, CurrLineLen())
    GotoBufferId(org_id)
    return(result)
  end StrReplace
#endif


// End of compatibility restrictions and mitigations





// START OF GET_DRIVE_TYPE() IMPLEMENTATION


/*
  Include file    get_drive_type.inc
  Author          Carlo Hogeveen
  Website         ecarlo.nl/tse
  Version         v2   9 Nov 2023
  Compatibility   TSE 3.0 upwards, all variants

  This include file provides the function get_drive_type().

  Unlike the Windows function GetDriveType(), which only accepts root(drive)s,
  get_drive_type() also accepts directory and file paths provided they start
  with their root (drive).

  get_drive_type() accepts directories and files on shares,
  and likewise determines their roots and "drive types".
  For example: "\\Share\", "\\Server\Share", "\\Server\Share\secrets.txt".

  An intentional functional difference:
    Windows' GetDriveType("\\") returns DRIVE_REMOTE.
    This   get_drive_type("\\") returns DRIVE_UNKNOWN.
  For get_drive_type() it is more useful to treat "\\" as an invalid root,
  because otherwise non-existent shares would appear to exist on the remote
  drive "\\".

  get_drive_type() is optimized for speed:
  Previously determined roots and their types are recognized in and applied to
  subsequend paths without having to consult GetDriveType() again.

  In Linux get_drive_type() always returns DRIVE_FIXED, pretending everything
  is on a fixed harddrive.


  INSTALLATION

  It is possible to put this file in the same directory as your own macro,
  and to include the file into your own macro with the line:
    #include ["get_drive_type.inc"]

  My preferred way of installing include files is to copy their content into my
  own macros.
  This avoids the possibility that a version of a macro might not match with a
  version of an include file.


  HISTORY

  v1   11 Mar 2021
    This version was based on slow Dos() calls to "WMIC",
    and was only compatible with Windows TSE 4.0 upwards.

  v2   9 Nov 2023
    For Windows this version uses a fast Windows API,
    and is itself optimized for speed.

    For Linux it simplisticly reports every path as being on a fixed drive.
    This makes macros that use get_drive_type() "compatible" with Linux.

*/





/*

#define DRIVE_UNKNOWN     0 // The drive type cannot be determined.
#define DRIVE_NO_ROOT_DIR 1 // The root path is invalid; for example, there is no volume mounted at the specified path.
#define DRIVE_REMOVABLE   2 // The drive has removable media; for example, a floppy drive, thumb drive, or flash card reader.
#define DRIVE_FIXED       3 // The drive has fixed media; for example, a hard disk drive or flash drive.
#define DRIVE_REMOTE      4 // The drive is a remote (network) drive.
#define DRIVE_CDROM       5 // The drive is a CD-ROM drive.
#define DRIVE_RAMDISK     6 // The drive is a RAM disk.


#ifdef LINUX

  integer proc get_drive_type(string fqn_string)
    string dummy [1] = fqn_string
    dummy = dummy // To avoid compiler warning.
    return(DRIVE_FIXED)
  end get_drive_type

#else

  dll "<kernel32.dll>"
    integer proc GetDriveType(
        string      lpRootPathName              : CStrVal
    )
    : "GetDriveTypeA"
  end

  integer get_drive_type_id      = 0
  integer get_drive_type_entries = 0

  integer proc get_drive_type(string fqn_string)
    integer fqn_type                   = DRIVE_UNKNOWN
    integer root_entry                 = 0
    string  root_string [MAXSTRINGLEN] = ''

    if get_drive_type_id == 0
      PushLocation()
      get_drive_type_id = CreateTempBuffer()
      ChangeCurrFilename(SplitPath(CurrMacroFilename(), _NAME_) + ':id',
                         _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      PopLocation()
    else
      for root_entry = 1 to get_drive_type_entries
        root_string = GetBufferStr('String' + Str(root_entry), get_drive_type_id)
        if Pos(root_string, fqn_string) == 1
          fqn_type = GetBufferInt('Type' + Str(root_entry), get_drive_type_id)
          break
        endif
      endfor
    endif

    //  Note:
    //    I noticed one network reference needing a trailing slash,
    //    so parent paths are tested with and without it,
    //    trying the longer, more specific path (with traing slash) first.

    root_string = fqn_string

    while (fqn_type in DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR)
    and   Length(root_string) > 0
      fqn_type = GetDriveType(root_string)
      if (fqn_type in DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR)
        root_string = RemoveTrailingSlash(root_string)
        fqn_type    = GetDriveType(root_string)
      endif
      if (fqn_type in DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR)
        if Pos('\', root_string)
          root_string = SplitPath(root_string, _DRIVE_|_PATH_)
        else
          root_string = ''
        endif
      else
        // GetDriveType("\\") returns drive type DRIVE_REMOTE, which might be
        // an OK Windows default, but in TSE's context it is not a valid drive.
        if root_string == '\\'
          fqn_type    = DRIVE_UNKNOWN
          root_string = ''
        else
          get_drive_type_entries = get_drive_type_entries + 1
          SetBufferStr('String' + Str(get_drive_type_entries), root_string,
                       get_drive_type_id)

          SetBufferInt('Type'   + Str(get_drive_type_entries), fqn_type,
                       get_drive_type_id)
        endif
      endif
    endwhile

    return(fqn_type)
  end get_drive_type

#endif

*/


// END OF GET_FRIVE_TYPE() IMPLEMENTATION





// START OF BIGINT IMPLEMENTATION

forward string  proc add_bigints (string term1, string term2)
forward string  proc subtract_bigint_minus_bigint (string minuend,
                                                   string subtrahend)
forward string  proc multiply_bigints (string factor1, string factor2)
forward string  proc divide_bigint_by_bigint (string dividend, string divisor)
forward integer proc compare_bigints
                                  (string left_quantity, string right_quantity)


// START OF INTERNAL FUNCTIONS - DO NOT CALL THESE

proc standardize_bigint(    string  input_bigint,
                        var string  output_positive_bigint,
                        var integer output_bigint_is_negative)
  integer i = 0
  // Remove surrounding white space and initialize the output sign as positive.
  output_positive_bigint    = Trim(input_bigint)
  output_bigint_is_negative = FALSE
  // Remove leading signs and remaining leading spaces,
  // toggling the output sign for each minus sign.
  while (output_positive_bigint[1] in '+', '-', ' ')
    if output_positive_bigint[1] == '-'
      output_bigint_is_negative = not output_bigint_is_negative
    endif
    output_positive_bigint = output_positive_bigint[2:MAXSTRINGLEN]
  endwhile
  // Remove trailing signs and remaining trailing spaces,
  // toggling the output sign for each minus sign.
  while (   output_positive_bigint[Length(output_positive_bigint)]
         in '+', '-', ' ')
    if output_positive_bigint[Length(output_positive_bigint)] == '-'
      output_bigint_is_negative = not output_bigint_is_negative
    endif
    output_positive_bigint =
      output_positive_bigint[1: Length(output_positive_bigint) - 1]
  endwhile
  // Delete group separators, like thousands separators, but deleting them
  // anywhere between two digits to account for all countries and styles.
  // Any decimal separators (commas or points) are not recognized as such and
  // are deleted when they resemble group separators.
  i = 2
  while i < Length(output_positive_bigint)
    if  (output_positive_bigint[i    ] in ',', '.', ' ', "'", '_')
    and (output_positive_bigint[i - 1] in '0' .. '9'             )
    and (output_positive_bigint[i + 1] in '0' .. '9'             )
      output_positive_bigint = output_positive_bigint[1     : i - 1       ]
                             + output_positive_bigint[i + 1 : MAXSTRINGLEN]
    else
      i = i + 1
    endif
  endwhile
  // Set output to an empty string if an illegal character is found.
  for i = 1 to Length(output_positive_bigint)
    if not (output_positive_bigint[i] in '0' .. '9')
      output_positive_bigint = ''
      break
    endif
  endfor
  // Remove leading zeros except the last one if the output value is zero.
  while output_positive_bigint[1] == '0'
  and   Length(output_positive_bigint) > 1
    output_positive_bigint = output_positive_bigint[2:MAXSTRINGLEN]
  endwhile
  // If the output value is zero, then make it not negative.
  if output_positive_bigint == '0'
    output_bigint_is_negative = FALSE
  endif
end standardize_bigint

integer proc compare_positive_bigints(string left_quantity, string right_quantity)
  integer result = 0
  if Length(left_quantity) > Length(right_quantity)
    result = 1
  elseif Length(left_quantity) < Length(right_quantity)
    result = -1
  elseif left_quantity > right_quantity
    result = 1
  elseif left_quantity < right_quantity
    result = -1
  endif
  return(result)
end compare_positive_bigints

string proc add_positive_bigints(string p_term1, string p_term2)
  integer carry                       = 0
  integer i                           = 0
  integer last_digit                  = 0
  integer len                         = 0
  integer sum_of_two_digits_and_carry = 0
  string  term1        [MAXSTRINGLEN] = p_term1
  string  term2        [MAXSTRINGLEN] = p_term2
  string  total_sum    [MAXSTRINGLEN] = ''
  if  p_term1 <> ''
  and p_term2 <> ''
    if Length(term1) > Length(term2)
      len   = Length(term1)
      term2 = Format(term2:len:'0')
    else
      len   = Length(term2)
      term1 = Format(term1:len:'0')
    endif
    for i = len downto 1
      sum_of_two_digits_and_carry = Val(term1[i]) + Val(term2[i]) + carry
      last_digit                  = sum_of_two_digits_and_carry mod 10
      carry                       = sum_of_two_digits_and_carry  /  10
      total_sum                   = Str(last_digit) + total_sum
    endfor
    if carry
      if Length(total_sum) == MAXSTRINGLEN
        total_sum = ''
      else
        total_sum = Str(carry) + total_sum
      endif
    endif
  endif
  return(total_sum)
end add_positive_bigints

string proc subtract_positive_larger_minus_smaller_bigint(string p_minuend,
                                                          string p_subtrahend)
  string  minuend      [MAXSTRINGLEN] = p_minuend
  string  subtrahend      [MAXSTRINGLEN] = p_subtrahend
  integer i                     = 0
  integer j                     = 0
  integer len                   = 0
  integer keep_borrowing        = FALSE
  string  difference [MAXSTRINGLEN] = ''
  if  p_minuend    <> ''
  and p_subtrahend <> ''
    if Length(minuend) > Length(subtrahend)
      len        = Length(minuend)
      subtrahend = Format(subtrahend:len:'0')
    else
      len     = Length(subtrahend)
      minuend = Format(minuend:len:'0')
    endif
    for i = len downto 1 by 1
      if     minuend[i] > subtrahend[i]
        difference     = Str(Val(minuend[i]) - Val(subtrahend[i]))
                       + difference
      elseif minuend[i] < subtrahend[i]
        difference     = Str(Val(minuend[i]) + 10 - Val(subtrahend[i]))
                       + difference
        j              = i - 1
        keep_borrowing = TRUE
        while j > 0
        and   keep_borrowing
          if minuend[j] == '0'
            minuend[j]     = '9'
          else
            minuend[j]     = Str(Val(minuend[j]) - 1)
            keep_borrowing = FALSE
          endif
          j = j - 1
        endwhile
      else
        difference = '0' + difference
      endif
    endfor
    while difference[1]     == '0'
    and   Length(difference) >  1
      difference = difference[2:MAXSTRINGLEN]
    endwhile
  endif
  return(difference)
end subtract_positive_larger_minus_smaller_bigint

/*
  For multiplying I use an algorith of my own that keeps intermediate results
  small. This turns out to be ideal for TSE's small integer size.

  It is based on this principle:
  Call each right-most digit a ones-digit, each digit to the left of that a
  tens-digit, each digit to the left of that a hundreds-digit, etcetera.
  For this example assume two three-digit factors to multiply to a product.
  The ones-digit of the product is only determined by the multiplication of the
  ones-digits of the two factors, and also results in a carry.
  The tens-digit of the product is determined by adding the carry plus the
  multiplications of the ones-digits of one factor with the tens-digits
  of the other factor, again resulting in a new carry.
  The hundreds-digit of the product is determined by adding the carry plus the
  multiplications of the ones-digits of one factor with the hundreds-digits
  of the other factor plus the multiplication of the tens-digits of both
  factors, again resulting in a new carry.
  The thousands-digit of the product is determined by adding the carry plus the
  multiplications of the tens-digit of one factor with the hundreds-digit of
  the other factor, again resulting in a new carry.
  The ten-thousands-digit of the product is determined by adding the carry plus
  the multiplication of the hundreds-digits of the factors, again resulting in
  a new carry.
  The hundred-thousands-digit of the product is determined by the final carry.

  For example
      123
      456 *
      ---
    56088
  can be calculated as follows
        6 * 3                 = 8 carry 1
    1 + 6 * 2 + 5 * 3         = 8 carry 2
    2 + 6 * 1 + 5 * 2 + 4 * 3 = 0 carry 3
    3 + 5 * 1 + 4 * 2         = 6 carry 1
    1 + 4 * 1                 = 5 carry 0
*/

string proc multiply_positive_bigints(string p_factor1, string p_factor2)
  integer carry                  = 0
  string  factor1 [MAXSTRINGLEN] = p_factor1
  string  factor2 [MAXSTRINGLEN] = p_factor2
  integer factors_left_pos       = 0
  integer factors_right_pos      = 0
  integer factor1_pos            = 0
  integer factor2_pos            = 0
  integer len                    = 0
  string  product [MAXSTRINGLEN] = ''
  integer product_pos_digit      = 0
  integer product_pos_product    = 0
  integer size_error             = FALSE
  if Length(factor1) > Length(factor2)
    len     = Length(factor1)
    factor2 = Format(factor2:len:'0')
  else
    len     = Length(factor2)
    factor1 = Format(factor1:len:'0')
  endif
  factors_left_pos = len
  for factors_right_pos = len downto 1 by 1
    for factors_left_pos = factors_left_pos downto 1
      if Length(product) == MAXSTRINGLEN
        size_error = TRUE
      else
        product_pos_product = 0
        for factor1_pos = factors_right_pos downto factors_left_pos
          factor2_pos         = factors_left_pos
                              + (factors_right_pos - factor1_pos)
          product_pos_product = product_pos_product
                              + Val(factor1[factor1_pos])
                              * Val(factor2[factor2_pos])
        endfor
        product_pos_product = product_pos_product + carry
        carry               = product_pos_product / 10
        product_pos_digit   = product_pos_product mod 10
        product             = Str(product_pos_digit) + product
      endif
    endfor
    factors_left_pos = 1
  endfor
  if carry
    if Length(product) == MAXSTRINGLEN
      size_error = TRUE
    else
      product = Str(carry) + product
    endif
  endif
  if size_error
    product = ''
  endif
  while product[1]     == '0'
  and   Length(product) >  1
    product = product[2:MAXSTRINGLEN]
  endwhile
  return(product)
end multiply_positive_bigints

/*
  The algorithm used is ordinary long division.
  Division is an expensive operation.
  A small performance improvement is realized by storing the results of some
  multiplications that are likely to recur for the same division.
*/

string bigint_pre_multiplied_divisor [MAXSTRINGLEN] = ''
string bigint_divisor_times_0        [MAXSTRINGLEN] = ''
string bigint_divisor_times_1        [MAXSTRINGLEN] = ''
string bigint_divisor_times_2        [MAXSTRINGLEN] = ''
string bigint_divisor_times_3        [MAXSTRINGLEN] = ''
string bigint_divisor_times_4        [MAXSTRINGLEN] = ''
string bigint_divisor_times_5        [MAXSTRINGLEN] = ''
string bigint_divisor_times_6        [MAXSTRINGLEN] = ''
string bigint_divisor_times_7        [MAXSTRINGLEN] = ''
string bigint_divisor_times_8        [MAXSTRINGLEN] = ''
string bigint_divisor_times_9        [MAXSTRINGLEN] = ''
string bigint_divisor_times_10       [MAXSTRINGLEN] = ''

string proc get_bigint_divisor_product(string  p_divisor,
                                       integer p_number_of_times)
  string product [MAXSTRINGLEN] = ''
  if p_divisor <> bigint_pre_multiplied_divisor
    bigint_pre_multiplied_divisor = p_divisor
    bigint_divisor_times_0        = '0'
    bigint_divisor_times_1        = p_divisor
    bigint_divisor_times_2        = ''
    bigint_divisor_times_3        = ''
    bigint_divisor_times_4        = ''
    bigint_divisor_times_5        = ''
    bigint_divisor_times_6        = ''
    bigint_divisor_times_7        = ''
    bigint_divisor_times_8        = ''
    bigint_divisor_times_9        = ''
    if Length(p_divisor) <> MAXSTRINGLEN
      bigint_divisor_times_10     = p_divisor + '0'
    endif
  endif
  case p_number_of_times
    when  0 product = bigint_divisor_times_0
    when  1 product = bigint_divisor_times_1
    when  2 product = bigint_divisor_times_2
    when  3 product = bigint_divisor_times_3
    when  4 product = bigint_divisor_times_4
    when  5 product = bigint_divisor_times_5
    when  6 product = bigint_divisor_times_6
    when  7 product = bigint_divisor_times_7
    when  8 product = bigint_divisor_times_8
    when  9 product = bigint_divisor_times_9
    when 10 product = bigint_divisor_times_10
  endcase
  if product == ''
    product = multiply_positive_bigints(p_divisor, Str(p_number_of_times))
    case p_number_of_times
      when  0 bigint_divisor_times_0  = product
      when  1 bigint_divisor_times_1  = product
      when  2 bigint_divisor_times_2  = product
      when  3 bigint_divisor_times_3  = product
      when  4 bigint_divisor_times_4  = product
      when  5 bigint_divisor_times_5  = product
      when  6 bigint_divisor_times_6  = product
      when  7 bigint_divisor_times_7  = product
      when  8 bigint_divisor_times_8  = product
      when  9 bigint_divisor_times_9  = product
      when 10 bigint_divisor_times_10 = product
    endcase
  endif
  return(product)
end get_bigint_divisor_product

string proc divide_positive_bigint_by_bigint(string p_dividend,
                                             string p_divisor)
  integer dividend_digit_index                     = 0
  string  dividend_part        [MAXSTRINGLEN] = ''
  string  divisor_product      [MAXSTRINGLEN] = ''
  integer number_of_times                     = 0
  string  prev_divisor_product [MAXSTRINGLEN] = ''
  string  quotient             [MAXSTRINGLEN] = ''
  for dividend_digit_index = 1 to Length(p_dividend)
    dividend_part   = iif(dividend_part == '0', '', dividend_part)
                    + p_dividend[dividend_digit_index]
    number_of_times =  0
    divisor_product = '0'
    repeat
      number_of_times      = number_of_times + 1
      prev_divisor_product = divisor_product
      divisor_product      = get_bigint_divisor_product(p_divisor,
                                                        number_of_times)
    until number_of_times                                          == 10
       or compare_positive_bigints(divisor_product, dividend_part) ==  1
    dividend_part = subtract_positive_larger_minus_smaller_bigint(
                                          dividend_part, prev_divisor_product)
    quotient      = quotient + Str(number_of_times - 1)
  endfor
  while quotient[1]     == '0'
  and   Length(quotient) >  1
    quotient = quotient[2:MAXSTRINGLEN]
  endwhile
  return(quotient)
end divide_positive_bigint_by_bigint

proc bigint_compiler_pacifier()
  integer i    = 0
  string  s[1] = ''
  if FALSE
    bigint_compiler_pacifier()
    standardize_bigint(s, s, i)
    compare_positive_bigints(s, s)
    add_positive_bigints(s, s)
    subtract_positive_larger_minus_smaller_bigint(s, s)
    multiply_positive_bigints(s, s)
    get_bigint_divisor_product(s, i)
    divide_positive_bigint_by_bigint(s, s)

    add_bigints(s, s)
    subtract_bigint_minus_bigint(s, s)
    multiply_bigints(s, s)
    divide_bigint_by_bigint(s, s)
    compare_bigints(s, s)
  endif
end bigint_compiler_pacifier

// END OF INTERNAL FUNCTIONS - DO NOT CALL THESE





// START OF SAFE FUNCTIONS - THESE CAN BE CALLED BY YOUR OWN CODE

string proc add_bigints(string p_term1, string p_term2)
  string  sum   [MAXSTRINGLEN] = ''
  string  term1 [MAXSTRINGLEN] = ''
  integer term1_is_negative    = FALSE
  string  term2 [MAXSTRINGLEN] = ''
  integer term2_is_negative    = FALSE
  standardize_bigint(p_term1, term1, term1_is_negative)
  standardize_bigint(p_term2, term2, term2_is_negative)
  if  term1 <> ''
  and term2 <> ''
    if term1_is_negative == term2_is_negative
      sum = add_positive_bigints(term1, term2)
      if  sum <> ''
      and term1_is_negative
        sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
      endif
    else
      case compare_positive_bigints(term1, term2)
        when 0
          sum = '0'
        when -1
          sum = subtract_positive_larger_minus_smaller_bigint(term2, term1)
          if  sum <> ''
          and term2_is_negative
            sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
          endif
        when 1
          sum = subtract_positive_larger_minus_smaller_bigint(term1, term2)
          if  sum <> ''
          and term1_is_negative
            sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
          endif
      endcase
    endif
  endif
  return(sum)
end add_bigints

string proc subtract_bigint_minus_bigint(string p_minuend, string p_subtrahend)
  string  difference [MAXSTRINGLEN] = ''
  string  minuend    [MAXSTRINGLEN] = ''
  integer minuend_is_negative       = FALSE
  string  subtrahend [MAXSTRINGLEN] = ''
  integer subtrahend_is_negative    = FALSE
  standardize_bigint(p_minuend   , minuend   ,    minuend_is_negative)
  standardize_bigint(p_subtrahend, subtrahend, subtrahend_is_negative)
  if  minuend    <> ''
  and subtrahend <> ''
    if Length(Trim(subtrahend)) == MAXSTRINGLEN
      difference = ''
    else
      difference = add_bigints(minuend, '-' + subtrahend)
    endif
  endif
  return(difference)
end subtract_bigint_minus_bigint

string proc multiply_bigints(string p_factor1, string p_factor2)
  string  factor1 [MAXSTRINGLEN] = ''
  integer factor1_is_negative    = FALSE
  string  factor2 [MAXSTRINGLEN] = ''
  integer factor2_is_negative    = FALSE
  string  product [MAXSTRINGLEN] = ''
  standardize_bigint(p_factor1, factor1, factor1_is_negative)
  standardize_bigint(p_factor2, factor2, factor2_is_negative)
  if  factor1 <> ''
  and factor2 <> ''
    product = multiply_positive_bigints(factor1, factor2)
    if  product             <> ''
    and factor1_is_negative <> factor2_is_negative
      product = iif(Length(product) == MAXSTRINGLEN, '', '-' + product)
    endif
  endif
  return(product)
end multiply_bigints

string proc divide_bigint_by_bigint(string p_dividend, string p_divisor)
  string  dividend [MAXSTRINGLEN] = ''
  integer dividend_is_negative    = FALSE
  string  divisor  [MAXSTRINGLEN] = ''
  integer divisor_is_negative     = FALSE
  string  quotient [MAXSTRINGLEN] = ''
  standardize_bigint(p_dividend, dividend, dividend_is_negative)
  standardize_bigint(p_divisor , divisor , divisor_is_negative )
  if  dividend <> ''
  and divisor  <> ''
  and divisor  <> '0'
    quotient = divide_positive_bigint_by_bigint(dividend, divisor)
    if  quotient             <> ''
    and dividend_is_negative <> divisor_is_negative
      quotient = iif(Length(quotient) == MAXSTRINGLEN, '', '-' + quotient)
      quotient = iif(quotient == '-0', '0', quotient)
    endif
  endif
  return(quotient)
end divide_bigint_by_bigint

integer proc compare_bigints(string p_left_quantity, string p_right_quantity)
  integer comparison                    = MININT
  string  left_quantity  [MAXSTRINGLEN] = ''
  integer left_quantity_is_negative     = FALSE
  string  right_quantity [MAXSTRINGLEN] = ''
  integer right_quantity_is_negative    = FALSE
  standardize_bigint(p_left_quantity , left_quantity , left_quantity_is_negative )
  standardize_bigint(p_right_quantity, right_quantity, right_quantity_is_negative)
  if  left_quantity  <> ''
  and right_quantity <> ''
    if left_quantity_is_negative == right_quantity_is_negative
      comparison = compare_positive_bigints(left_quantity, right_quantity)
      comparison = iif(left_quantity_is_negative, comparison * -1, comparison)
    else
      comparison = iif(left_quantity_is_negative, -1, 1)
    endif
  endif
  return(comparison)
end compare_bigints

// END OF SAFE FUNCTIONS - THESE CAN BE CALLED BY YOUR OWN CODE

// END OF BIGINT IMPLEMENTATION





// Global constants and semi-constants
#define ALL_FILES                   -1
string  FS_CHARCTERS         [32] = ChrSet('~\d000\d010\d013\d032') // Intentionally overly broad.
#define INVALID_HANDLE              -1
string  MACRO_NAME [MAXSTRINGLEN] = ''


// Global variables
integer ads_option             = TRUE
integer calculate_dir_sizes    = TRUE
integer dir_option             = TRUE
integer file_option            = TRUE
integer num_dirs               = 0
integer num_files              = 0
integer num_inaccessible_dirs  = 0
integer num_inaccessible_files = 0

integer pos_parent_line        = 1   // Relative positions
integer pos_group_size         = 2
integer pos_modified_date      = 3
integer pos_modified_time      = 4
integer pos_size               = 5
integer pos_accessibility      = 6
integer pos_type               = 7
integer pos_name               = 8


proc to_beep_or_not_to_beep()
  if Query(Beep)
    Alarm()
  endif
end to_beep_or_not_to_beep


integer proc msg(string s)
  integer ok = TRUE

  if Query(MsgLevel) == _ALL_MESSAGES_
    KeyPressed()
    Message(MACRO_NAME, ':';; s;; '... ')
  endif

  if KeyPressed()
    if Query(Key) == <Escape>
      Warn(MACRO_NAME, ': You escaped')
      ok = FALSE
    endif
  endif

  return(ok)
end msg


string proc group_digits(string s)
  integer i                = 0
  string  r [MAXSTRINGLEN] = s

  for i = Length(r) - 3 downto 1 by 3
    if r[i] <> ''
      r = r[1: i] + ',' + r[i + 1: MAXSTRINGLEN]
    endif
  endfor

  return(r)
end group_digits


string proc get_system_date_time()
  string hh_mm_ss         [8] = GetTimeStr()
  string yyyymmdd_hhmmss [15] = ''
  string yyyy_mm_dd      [10] = GetDateStr()
  yyyymmdd_hhmmss = yyyy_mm_dd[1:4] + yyyy_mm_dd[6:2] + yyyy_mm_dd[9:2] +
                    '_' +
                    hh_mm_ss[1:2] + hh_mm_ss[4:2] + hh_mm_ss[7:2]
  yyyymmdd_hhmmss = StrReplace(' ', yyyymmdd_hhmmss, '0', '')
  return(yyyymmdd_hhmmss)
end get_system_date_time


string proc get_field(integer field_number)
  string field_value [MAXSTRINGLEN] = ''
  PushLocation()
  BegLine()
  WordRight(field_number - 1)
  if field_number == pos_name
    field_value = GetText(CurrPos(), MAXSTRINGLEN)
  else
    field_value = GetWord()
  endif
  PopLocation()
  return(field_value)
end get_field


proc set_field(integer field_number, string new_value)
  BegLine()
  WordRight(field_number - 1)
  MarkWord()
  KillBlock()
  InsertText(new_value)
end set_field


proc add_to_object_and_ancestors(string additional_group_size)
  string  group_size [MAXSTRINGLEN] = ''
  integer parent_line               = CurrLine()

  PushLocation()

  repeat
    GotoLine(parent_line)
    group_size  = get_field(pos_group_size)
    group_size  = add_bigints(group_size, additional_group_size)
    set_field(pos_group_size, group_size)
    parent_line = Val(get_field(pos_parent_line))
  until parent_line == 0

  PopLocation()
end add_to_object_and_ancestors


integer proc dir_list(string dir_names)
  integer check_line                = 1
  integer dir_counter               = 1
  string  dir_name   [MAXSTRINGLEN] = ''
  string  found_type            [1] = ''
  string  group_size [MAXSTRINGLEN] = ''
  integer handle                    = INVALID_HANDLE
  integer num_messages              = 0
  integer num_objects               = 0
  integer num_time_checks           = 0
  integer ok                        = TRUE
  integer prev_num_messages         = 0

  // Create and name the empy list.
  CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + '_' + get_system_date_time() + '.txt',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)

  // Put the input directories in the list.
  for dir_counter = 1 to NumFileTokens(dir_names)
    dir_name    = GetFileToken(dir_names, dir_counter)

    if  Length(dir_name) == 2
    and isAlpha(dir_name[1])
    and dir_name[2] == ':'
      dir_name = AddTrailingSlash(dir_name)
    endif

    //  The name does not have a fixed length, and is the only field that might
    //  contain spaces, so we put it last.
    //  The dir size does not get a fixed length, because it would need to be
    //  too big.
    //  Because we have to do formatting later anyway, at this stage we do not
    //  use fixed Length fields.
    //  Fields: parent_line, group_size, type, date, time, own_size, name.
    FindThisFile(dir_name)
    AddLine(dir_name)
    BegLine()
    // Initial set_field() for parameter directory.
    InsertText(Format(0; 0; FFDateStr(); FFTimeStr(); 0; 1; 'd', ' '))
  endfor

  //  For the first to the last line of the list:
  //    If the line refers to a directory,
  //    then add its direct members at lines directly below it,
  BegFile()
  while ok
  and   check_line <= NumLines()

    //  Messages slow the tool down, so only give them sparsely.
    // At most once every 1000 files AND at most once a second.
    num_objects = num_objects + 1
    if num_objects / 1000 <> num_time_checks
      num_time_checks = num_objects / 1000
      num_messages = GetTime() / 100
      if num_messages <> prev_num_messages
        prev_num_messages = num_messages
        ok = msg(Format('Generating:';;
                         // num_messages;;
                         group_digits(Str(num_objects));
                         'objects';;
                         get_field(pos_name)))
      endif
    endif

    GotoLine(check_line)

    if get_field(pos_type) == 'd'
      if Val(get_field(pos_accessibility))
        num_dirs = num_dirs + 1
        dir_name = get_field(pos_name)

        if Length(dir_name) <= MAXSTRINGLEN - 1
          dir_name = AddTrailingSlash(dir_name)
          handle   = FindFirstFile(dir_name + '*', ALL_FILES)
          if handle <> INVALID_HANDLE
            group_size = '0'
            repeat
              if not (FFName() in '.', '..')
                found_type  = iif(FFAttribute() & _DIRECTORY_, 'd', 'f')
                AddLine(FFName())
                BegLine()
                InsertText(Format(check_line;   // Parent line.
                                  0;            // Default group size.
                                  FFDateStr();
                                  FFTimeStr();
                                  FFSizeStr();
                                  '1';          // Assume accessible.
                                  found_type,   // Dir or file.
                                  ' '))
                InsertText(dir_name)

                if found_type == 'f'
                  if Length(dir_name) + Length(FFName()) > MAXSTRINGLEN
                  or not FileExists(dir_name + FFName())
                    EndLine()
                    set_field(pos_accessibility, '0')
                    num_inaccessible_files = num_inaccessible_files + 1
                  endif

                  if  dir_option
                  and calculate_dir_sizes
                    group_size = add_bigints(group_size, FFSizeStr())
                  endif
                else
                  if Length(dir_name) + Length(FFName()) + 2 > MAXSTRINGLEN
                  or not FileExists(dir_name + FFName())
                    EndLine()
                    set_field(pos_accessibility, '0')
                    num_inaccessible_dirs = num_inaccessible_dirs + 1
                  endif
                endif
              endif
            until not FindNextFile(handle, ALL_FILES)

            FindFileClose(handle)
            GotoLine(check_line)

            if group_size <> '0'
              add_to_object_and_ancestors(group_size)
            endif
          endif
        endif
      endif
    else
      num_files = num_files + 1
    endif
    check_line = check_line + 1
  endwhile

  BufferType(_NORMAL_)
  return(ok)
end dir_list


integer proc copy_group_to_dir_sizes()
  integer ok = TRUE

  BegFile()

  repeat
    if get_field(pos_type) == 'd'
      set_field(pos_size, get_field(pos_group_size))
    endif
  until not Down()

  return(ok)
end copy_group_to_dir_sizes


integer proc format_list()
  integer largest_digit_grouped_size = 0
  integer largest_object_size        = 0
  integer more_lines                 = TRUE
  integer ok                         = TRUE

  // Delete the lines of excluded object types.
  if (not dir_option or not file_option)
    ok = msg('Formatting:  Deleting temporary objects')
    BegFile()
    while more_lines
      case get_field(pos_type)
        when 'f'
          if file_option
            more_lines = Down()
          else
            more_lines = (CurrLine() < NumLines())
            KillLine()
          endif
        when 'd'
          if dir_option
            more_lines = Down()
          else
            more_lines = (CurrLine() < NumLines())
            KillLine()
          endif
        otherwise
          ok = Warn(MACRO_NAME; 'program error 1.')
          more_lines = Down()
      endcase
    endwhile
  endif

  // Delete the two leading fields; parent line and group size.
  if  ok
  and NumLines()
    ok = msg('Formatting:  Deleting temporary columns')
    BegFile()
    repeat
      BegLine()
      WordRight(2)
      MarkColumn(CurrLine(), 1, CurrLine(), CurrCol() - 1)
      KillBlock()
    until not Down()

    // Adjust the fields' relative positions
    pos_modified_date = pos_modified_date - 2
    pos_modified_time = pos_modified_time - 2
    pos_size          = pos_size          - 2
    pos_accessibility = pos_accessibility - 2
    pos_type          = pos_type          - 2
    pos_name          = pos_name          - 2

    // Determine the largest object size.
    ok = msg('Formatting:  Determining the largest object size')
    BegFile()
    repeat
      if Length(get_field(pos_size)) > largest_object_size
        largest_object_size = Length(get_field(pos_size))
      endif
    until not Down()

    // Digit-group and right-justify object sizes based on the largest.
    ok = msg('Formatting:  Digit-grouping and right-adjusting object sizes')
    largest_digit_grouped_size = largest_object_size +
                                (largest_object_size - 1) / 3
    BegFile()
    repeat
      set_field(pos_size, Format(group_digits(get_field(pos_size)):
                                   largest_digit_grouped_size))
    until not Down()
  endif

  return(ok)
end format_list


integer proc sort_list()
  integer ok = TRUE
  integer old_MsgLevel = 0

  if NumLines() > 1
    // Sort on name
    old_MsgLevel = Set(MsgLevel, _WARNINGS_ONLY_)
    BegFile()
    WordRight(pos_name - 1)
    MarkColumn(1, CurrCol(), NumLines(), LongestLineInBuffer())
    ok = ExecMacro('sort -i')
    if not ok
      Warn(MACRO_NAME; 'abort: sorting on name failed.')
    endif
    Set(MsgLevel, old_MsgLevel)
  endif

  return(ok)
end sort_list


proc init_options()
  if  ads_option
  and dir_option
  and file_option
    ads_option  = FALSE
    dir_option  = FALSE
    file_option = FALSE
  endif
end init_options


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)
end WhenLoaded


proc Main()
  string  dir_names  [MAXSTRINGLEN] = ''
  integer ok                        = TRUE
  integer old_DateFormat            = Set(DateFormat, 6)           // Sortable.
  integer old_DateSeparator         = Set(DateSeparator, Asc('-')) // Not " ".
  integer old_Insert                = Set(Insert, ON)
  integer old_TimeFormat            = Query(TimeFormat)            // Sortable.
  integer old_TimeSeparator         = Set(TimeSeparator, Asc(':')) // Not " ".
  string  old_WordSet          [32] = Set(WordSet, FS_CHARCTERS)
  string  parameter  [MAXSTRINGLEN] = ''
  string  parameters [MAXSTRINGLEN] = Query(MacroCmdLine)
  integer param_counter             = 0
  integer run_time                  = 0
  integer start_time                = 0

  PushBlock()

  if (old_timeformat in 2, 4)
    old_timeformat = Set(TimeFormat, old_timeformat - 1)  // Sortable.
  endif

  if GetFileToken(parameters, 1) == ''
    ok = Ask('Which directories do you want to list:',
             parameters,
             GetFreeHistory(MACRO_NAME + ':Parameters'))
  endif

  start_time = GetTime()

  if      ok
  and not NumFileTokens(parameters)
    // Not ok to continue, but no error, so no warning.
    ok = FALSE
  endif

  if ok
    for param_counter = 1 to NumFileTokens(parameters)
      parameter = Trim(GetFileToken(parameters, param_counter))

      case Lower(parameter)
        when '-a'
          init_options()
          ads_option          = TRUE
        when '-d'
          init_options()
          dir_option    = TRUE
        when '-f'
          init_options()
          file_option         = TRUE
        when '--nodirsizes'
          calculate_dir_sizes = FALSE
        otherwise
          if FileExists(AddTrailingSlash(parameter)) & _DIRECTORY_
            dir_names = dir_names +
                        iif(dir_names == '', '', ' ') +
                        QuotePath(parameter)
          else
            to_beep_or_not_to_beep()
            ok = Warn(MACRO_NAME; 'abort: Input directory does not exist: "',
                      parameter, '".')
            break
          endif
      endcase
    endfor
  endif

  if  ok
  and dir_names  == ''
    // Not ok to continue, but not an error, so no warning.
    ok = FALSE
  endif

  ok = ok and msg('Generating')
  ok = ok and dir_list(dir_names)

  if  ok
  and dir_option
  and calculate_dir_sizes
    ok =        msg('Formatting:  Copying calculated directory sizes')
    ok = ok and copy_group_to_dir_sizes()
  endif

  ok = ok and msg('Formatting')
  ok = ok and format_list()
  ok = ok and msg('Sorting on full name')
  ok = ok and sort_list()

  PopBlock()
  BegFile()
  UpdateDisplay(_ALL_WINDOWS_REFRESH_)

  if ok
    Message(MACRO_NAME, ': Done.')
    to_beep_or_not_to_beep()

    run_time = GetTime()
    if run_time < start_time
      run_time = run_time + 24 * 60 * 60 * 100
    endif
    run_time = run_time - start_time

    MsgBox(MACRO_NAME,
           Format('Run-time:';; run_time / 6000,
                  ':',
                  (run_time mod 6000) / 100: 2: '0',
                  iif(run_time mod 100,
                      Format('.', run_time mod 100: 2: '0'),
                      ''),
                  Chr(13),
                  Chr(13),
                  iif(dir_option,
                      Format(group_digits(Str(num_dirs)); 'directories, of which';
                             group_digits(Str(num_inaccessible_dirs)); 'inaccessible.'),
                      ''),
                  iif(dir_option and file_option,
                      Chr(13),
                      ''),
                  iif(file_option,
                      Format(group_digits(Str(num_files)); 'files, of which';
                             group_digits(Str(num_inaccessible_files)); 'inaccessible.'),
                      '')))
  endif

  Set(DateFormat   , old_dateformat)
  Set(DateSeparator, old_DateSeparator)
  Set(Insert       , old_Insert )
  Set(TimeFormat   , old_TimeFormat)
  Set(TimeSeparator, old_TimeSeparator)
  Set(WordSet      , old_WordSet)
  PurgeMacro(CurrMacroFilename())
end Main

